type elem = content ref

and content =
  | Link of elem (* parent *)
  | Root of rank (* rank *)

and rank =
  int

let make () : elem = ref (Root 0);;
  
let find (x : elem) : elem =
  let rec find_root x = 
    match !x with
    | Link y -> find_root y
    | Root _ -> x
  and update_root r x =
    match !x with
    | Link y -> x := Link r; update_root r y
    | Root _ -> ()
  in
  let r = find_root x in
  update_root r x; r;;
      
let eq (x : elem) (y : elem) : bool =
  find x == find y;;
  
let link (x : elem) (y : elem) : unit =
  match !x, !y with
  | Root i, Root j ->
     if i > j then
       y := Link x
     else if i < j then
       x := Link y
     else begin
         x := Link y; y := Root (j+1)
       end
  | _ -> assert false;;
  
let union (x : elem) (y : elem) : unit =
  link (find x) (find y);;
